#include "WCSP.h"
#include "ycs_rcpsp.h"
#include <limits.h>
#include <string>
#include <iostream>



WCSP::WCSP() {
  totalWeight=0;
  contBStore=0;
  contWeithedYices=0;
  solver="yices";
  maximalCost=LONG_MAX;
  eqOne=false;
  typeSupport=0;
  nbVariables=0;
  initialCost=0;
  optimtotal=0;
 // level=0; 
  rcpspmax=false;
  verbose=false;
  ntasques=0;
  y=NULL;
}

void WCSP::setContext(yices_context & c,string ts,bool eqo,int sup,string codi,bool rcpspm,bool verb,ycs_rcpsp* b) {
  
  ctx=c;
  solver=ts;
  eqOne=eqo;
  typeSupport=sup;
  //level=l;  
  yices_set_arith_only(1);
  cod=codi;
  rcpspmax=rcpspm;
  verbose=verb;
  y=b;    
}

void WCSP::imprimir() {
  yices_dump_context(ctx);  
}


string ignore;
void pausa() {
    string paus;
    if(ignore.find("i")==std::string::npos) {
            cout << endl << "(paused, i=ignore next pauses)"<<endl;
        getline(cin, paus);
        ignore=paus;
    }
}

assertion_id WCSP::addConstraintRet(yices_expr e) {
  return yices_assert_retractable(ctx,e);
}
void WCSP::delConstraintRet(assertion_id id) {
  yices_retract(ctx,id);
}

void WCSP::addConstraint(yices_expr expr,long w) {
  //yices_pp_expr(expr); cout << endl;    
  if (w==0) {
    yices_assert(ctx,expr);
  } else {
    if (solver=="dicotomic") {addConstraintProtect(expr,w);}
    if (solver=="dico2") {addConstraintProtect(expr,w);}
    if (solver=="pm1") {addConstraintStore(expr,w);}   
    if (solver=="pm1.2") {addConstraintStore2(expr,w);}   
    if (solver=="pm1.3") {addConstraintStore2(expr,w);}   
    if (solver=="pm3") {addConstraintStore2(expr,w);}   
    if (solver=="yices") {addConstraintYices(expr,w);}          
  }
}

void WCSP::solve(bool model,int vMin,int vMax,map<string,yices_var_decl> & mapDeclVar) {   

 lbool lb=l_undef;
  //yices_dump_context(ctx);
  if (solver=="dicotomic") lb=dicotomic(model,vMin);   
  else if (solver=="dico2") lb=dico2(model,vMin);   
  else if (solver=="pm1") lb=pm1(1,0,vMin);   
  else if (solver=="pm1.2") lb=pm12(vMin);   
  else if (solver=="pm1.3") lb=pm13(vMin);   
  else if (solver=="pm3") lb=pm3();   
  else if (solver=="yices") lb=yices(); 
  else if (solver=="standard") lb=standard(model,vMin,vMax,mapDeclVar); 
  else if (solver=="linealup") lb=linealup(model,vMin,vMax,mapDeclVar); 
  else if (solver=="linealdown") lb=linealdown(model,vMin,vMax,mapDeclVar); 
  else if (solver=="standard2") lb=standard2(model,vMin,vMax,mapDeclVar); 
  else {cout << "Type solver <"<< solver << "> not supported" << endl;exit;}

  yices_model m = yices_get_model(ctx);
  string ms;
  yices_var_decl o;
  long s;

  switch(lb) {
  case l_true:
    if (solver=="dicotomic" || solver=="dico2" || solver=="pm1" || solver=="pm1.2" || solver=="pm1.3" || solver=="pm3") optimtotal=optimtotal+vMin; 
    if (verbose && (solver=="pm1" || solver=="pm1.2" || solver=="pm1.2" || solver=="pm1.2")) {
	yices_model m = yices_get_model(ctx);
        cout << "c ub " << optimtotal << endl;
        escriure_starts(m);
    }
    cout << "o " << optimtotal << endl;
    cout << "s OPTIMUM FOUND" << endl;
    if (model) y->mostraResultats();
    break;
  case l_false:
    cout<<"s UNSATISFIABLE"<<endl;
    break;
  case l_undef:   
    //cout<<"<<endl;
    break;
  } 
 if (d_debug) cout<<"Mètode Solving : "<<solver<<endl;        
}

void WCSP::setMaximalCost(long maxCost) {    
  maximalCost=maxCost;
}


void WCSP::setInitialCost(long initial) {
  initialCost=initial;
}


WCSP::~WCSP() {
}




///////////////////////////////////////////
//dico2
///////////////////////////////////////////

lbool WCSP::dico2(bool model,int vMin) {
  optim=0;
  if (vecConstraintsProtect.size()==0) return yices_check(ctx);  //There are not soft constraints
  if (typeSupport==1) maximalCost=maximalCost*2;
  if (totalWeight>maximalCost) totalWeight=maximalCost;
  //setting the bound 
  yices_var_decl vup = yices_mk_var_decl(ctx,(char *)"up",yices_mk_type(ctx,"int"));
  yices_assert(ctx,yices_mk_le(ctx,yices_mk_sum(ctx,& vecConstraintsProtect[0],vecConstraintsProtect.size()),yices_mk_var_from_decl(ctx,vup)));

  //yices_dump_context(ctx);  

  //check trivial upper bound
  yices_assert(ctx,yices_mk_le(ctx,yices_mk_var_from_decl(ctx,vup),yices_mk_num(ctx,totalWeight)));
  lbool status = yices_check(ctx);
  if(status == l_false) return l_false;
  yices_model m = yices_get_model(ctx);
  long upper_bound = totalWeight;
  yices_get_int_value(m,vup,&upper_bound);
  cout << "c ub "<< upper_bound+initialCost+vMin << endl;
  if (verbose) escriure_starts(m);
  //check trivial lower bound   
  long lower_bound = 0;
  if (upper_bound > 0) { 
    assertion_id vup_id = yices_assert_retractable(ctx,yices_mk_le(ctx,yices_mk_var_from_decl(ctx,vup),yices_mk_num(ctx, lower_bound)));	
    lbool status = yices_check(ctx);
    if(status == l_true) return l_true;
    yices_retract(ctx, vup_id);  
    cout << "c lb "<< 0+initialCost+vMin << endl;
  }else return l_true;

  long powtwo=1;
  do {
    assertion_id vup_id = yices_assert_retractable(ctx,yices_mk_le(ctx,yices_mk_var_from_decl(ctx,vup),yices_mk_num(ctx, powtwo)));
    status = yices_check(ctx);
    if (status==l_false) {
      yices_retract(ctx, vup_id);
      yices_assert(ctx,yices_mk_gt(ctx,yices_mk_var_from_decl(ctx,vup),yices_mk_num(ctx, powtwo)));
      lower_bound=powtwo;
      cout << "c lb "<< lower_bound+initialCost+vMin << endl;
      powtwo=powtwo*2;
    } else {
      upper_bound = powtwo;
      m = yices_get_model(ctx);
      yices_get_int_value(m,vup,&upper_bound);	
    }
  }while(status==l_false && powtwo<upper_bound);

  long op=upper_bound;
  
  do {
    long bound = (lower_bound+upper_bound)/2;
    optim=upper_bound;
    
    assertion_id vup_id = yices_assert_retractable(ctx,yices_mk_le(ctx,yices_mk_var_from_decl(ctx,vup),yices_mk_num(ctx, bound)));
    status = yices_check(ctx);
    if(status == l_false) {
      if (typeSupport==1) op=bound/2; else op=bound;
      cout << "c lb "<< op+initialCost+vMin << endl;
      lower_bound = bound;
      yices_retract(ctx, vup_id);  
      yices_assert(ctx,yices_mk_ge(ctx,yices_mk_var_from_decl(ctx,vup),yices_mk_num(ctx, bound+1)));
    }else if (status == l_true){
      upper_bound = bound;
      m = yices_get_model(ctx);
      yices_get_int_value(m,vup,&upper_bound);
       cout << "c ub "<< upper_bound+initialCost+vMin << endl;
       if (verbose) escriure_starts(m);
    }
  }while((lower_bound + 1)  < upper_bound );
  optimtotal=upper_bound;
  status=l_true;
  return status; 
}






///////////////////////////////////////////
//dicotomic
///////////////////////////////////////////

lbool WCSP::dicotomic(bool model,int vMin) {
  optim=0;
  if (vecConstraintsProtect.size()==0) return yices_check(ctx);  //There are not soft constraints
  if (typeSupport==1) maximalCost=maximalCost*2;
  if (totalWeight>maximalCost) totalWeight=maximalCost;
  //setting the bound 
  yices_var_decl vup = yices_mk_var_decl(ctx,(char *)"up",yices_mk_type(ctx,"int"));
  yices_assert(ctx,yices_mk_le(ctx,yices_mk_sum(ctx,& vecConstraintsProtect[0],vecConstraintsProtect.size()),yices_mk_var_from_decl(ctx,vup)));

  //yices_dump_context(ctx);  

  //check trivial upper bound
  yices_assert(ctx,yices_mk_le(ctx,yices_mk_var_from_decl(ctx,vup),yices_mk_num(ctx,totalWeight)));
  lbool status = yices_check(ctx);
  if(status == l_false) return l_false;
  yices_model m = yices_get_model(ctx);
  long upper_bound = totalWeight;
  yices_get_int_value(m,vup,&upper_bound);    
  long op2=upper_bound;
  if (typeSupport==1) op2=op2/2;
  cout << "c ub "<< op2+initialCost+vMin << endl;
  if (verbose) escriure_starts(m);
 
  //check trivial lower bound   
  long lower_bound = 0;
  if (upper_bound > 0) { 
    assertion_id vup_id = yices_assert_retractable(ctx,yices_mk_le(ctx,yices_mk_var_from_decl(ctx,vup),yices_mk_num(ctx, lower_bound)));	
    lbool status = yices_check(ctx);
    if(status == l_true) return l_true;
    yices_retract(ctx, vup_id);  
    cout << "c lb "<< 0+initialCost+vMin << endl;
  }else return l_true;

  long op=upper_bound;
  
  do {
    long bound = (lower_bound+upper_bound)/2;
    optim=upper_bound;
    
    assertion_id vup_id = yices_assert_retractable(ctx,yices_mk_le(ctx,yices_mk_var_from_decl(ctx,vup),yices_mk_num(ctx, bound)));
    status = yices_check(ctx);
    if(status == l_false) {
      if (typeSupport==1) op=bound/2; else op=bound;
      cout << "c lb "<< op+initialCost+vMin << endl;   
      lower_bound = bound;
      yices_retract(ctx, vup_id);  
      yices_assert(ctx,yices_mk_ge(ctx,yices_mk_var_from_decl(ctx,vup),yices_mk_num(ctx, bound+1)));
    }else if (status == l_true){
      upper_bound = bound;
      m = yices_get_model(ctx);
      yices_get_int_value(m,vup,&upper_bound);
      if (typeSupport==1) op=upper_bound/2; else op=upper_bound;
      cout << "c ub "<< op+initialCost+vMin << endl;
      if (verbose) escriure_starts(m);
    }
  }while((lower_bound + 1)  < upper_bound );
  optimtotal=upper_bound;
  status=l_true;
  return status; 
}

void WCSP::addConstraintProtect(yices_expr e,long w) {
  yices_expr argsor[2],terme[2];
  long act=vecConstraintsProtect.size()+1;
  char aux[20];

  //set the boolean variable b
  sprintf(aux,"b_%i",act);	  
  argsor[0]=e;
  argsor[1]=yices_mk_var_from_decl(ctx,yices_mk_var_decl(ctx,aux,yices_mk_type(ctx,"bool")));
  yices_assert(ctx,yices_mk_or(ctx,argsor,2));
  //set the the protect w*bi_act
  terme[0]=yices_mk_num(ctx,w); 
  terme[1]=reifVariable(argsor[1],act);
  vecConstraintsProtect.push_back(yices_mk_mul(ctx,terme,2));
  totalWeight=totalWeight+w;
}

///////////////////////////////////////////
//pm1.2
///////////////////////////////////////////

void WCSP::addConstraintStore2(yices_expr e,long w) {
  vecConstraints2.push_back(e); 
  vecWeigths2.push_back(w); 
  vecDown2.push_back(false); 
  setWeigths2.insert(w);
}

lbool WCSP::pm12(int vMin) {
  lbool lb=yices_check(ctx);
  long opt=0;
  long wMaxOld=LONG_MAX;
  if (lb!=l_true) return l_false;
  long wMax=0;
  for (set<long>::reverse_iterator its=setWeigths2.rbegin(); its!=setWeigths2.rend(); its++) {
    if (*its < wMaxOld && *its > wMax) {wMax=*its;break;}
  }
  while (wMax>0) {
    downConstraints(wMax);
    lb=pm1(2,wMax,vMin);
    opt=opt+optim;
    cout << "c lb " << (opt+initialCost+vMin) << endl;
    if (lb!=l_true) return l_false;
    wMaxOld=wMax;
    wMax=0;  
    for (set<long>::reverse_iterator its=setWeigths2.rbegin(); its!=setWeigths2.rend(); its++) {
      if (*its < wMaxOld && *its > wMax) {wMax=*its;break;}
    }
  }
  optimtotal=opt;
  //cout <<optimtotal << " " << opt << " " << optim << endl;
  return l_true;
}

lbool WCSP::downConstraints(long wMax) {
  for (int i=0;i<vecWeigths2.size();i++){
    if (vecWeigths2[i]>=wMax && !vecDown2[i]) {
      addConstraintStore(vecConstraints2[i],vecWeigths2[i]);
      vecDown2[i]=true;
    }
  }
}



///////////////////////////////////////////
//pm1.3
///////////////////////////////////////////



lbool WCSP::pm13(int vMin) {
  lbool lb=yices_check(ctx);
  long opt=0;
  long wMaxOld=0;
  if (lb!=l_true) return l_false;
  long wMax=LONG_MAX;
  for (set<long>::iterator its=setWeigths2.begin(); its!=setWeigths2.end(); its++) {
    if (*its > wMaxOld && *its < wMax) {wMax=*its;break;}
  }
  while (wMax<LONG_MAX) {
    downConstraints3(wMax);
    lb=pm1(2,wMax,vMin);
    opt=opt+optim;
    cout << "c lb " << opt+initialCost+vMin << endl;
    if (lb!=l_true) return l_false;
    wMaxOld=wMax;
    wMax=LONG_MAX;  
    for (set<long>::iterator its=setWeigths2.begin(); its!=setWeigths2.end(); its++) {
      if (*its > wMaxOld && *its < wMax) {wMax=*its;break;}
    }
  }
  optimtotal=opt;
  return l_true;
}

lbool WCSP::downConstraints3(long wMax) {
  for (int i=0;i<vecWeigths2.size();i++){
    if (vecWeigths2[i]<=wMax && !vecDown2[i]) {
      addConstraintStore(vecConstraints2[i],vecWeigths2[i]);
      vecDown2[i]=true;
    }
  }
}



///////////////////////////////////////////
//pm3
///////////////////////////////////////////

lbool WCSP::pm3() {
}


///////////////////////////////////////////
//pm1
///////////////////////////////////////////


lbool WCSP::pm1(int type,long wMax,int vMin) {
  long cost=0;
  optim=cost;
  lbool lb=yices_check(ctx);
  if (mapConstratintsStore.size()==0) return lb;  //There are not soft constraints
  while (lb==l_false) 
    {
      optim=cost;
      if (typeSupport==1) optim=optim/2;
      if (type<2) cout << "c lb " << (optim+initialCost+vMin) << endl;
      long ncore=yices_get_unsat_core_size(ctx);
      if (ncore==0) return l_false;
      assertion_id v[ncore];
      yices_get_unsat_core(ctx,v);
      long c=updatePm1(type,ncore,v,wMax); //make the proctections
      if (c==LONG_MAX) return l_false;
      cost=cost+c;
      lb=yices_check(ctx);
    }
  optimtotal=cost;
  optim=cost;
  return lb;
}



void WCSP::addConstraintStore(yices_expr e,long w) {
  assertion_id i=yices_assert_retractable(ctx,e); 
  regConstYices cy(i,e,w);
  mapConstratintsStore[i]=cy;	  
}



long WCSP::updatePm1(int type,long ncore,assertion_id v[],long wMax) {	
  long minor=LONG_MAX;
  for (long j=0;j<ncore;j++) {
    if (mapConstratintsStore.find(v[j])!=mapConstratintsStore.end()) if ((mapConstratintsStore.find(v[j])->second).getPes()<minor) minor=(mapConstratintsStore.find(v[j])->second).getPes();
  }
  
  vector<yices_expr> vecProtect;

  for (long i=0;i<ncore;i++) {
    if (mapConstratintsStore.find(v[i])!=mapConstratintsStore.end()) {
      //update the old constraints with w-minor
      regConstYices cyOld=mapConstratintsStore.find(v[i])->second;
      yices_expr args[2];
      args[0]=cyOld.getYicesExpr();
      long newWeight=cyOld.getPes()-minor;

      //create the new constraints with w=minor and guarded b_i
      char aux[20];
      sprintf(aux,"b_%i",contBStore);
      args[1]=yices_mk_var_from_decl(ctx,yices_mk_var_decl(ctx,aux,yices_mk_type(ctx,"bool")));
      if (eqOne) vecProtect.push_back(args[1]);
      else vecProtect.push_back(reifVariable(args[1],contBStore));
      contBStore++;
      yices_expr constProt=yices_mk_or(ctx,args,2);
      assertion_id j=yices_assert_retractable(ctx,constProt); 
      regConstYices cy(j,constProt,minor);
      mapConstratintsStore[j]=cy;

      //change de weight of the old constraints
      if (newWeight>0) {
	(mapConstratintsStore.find(v[i])->second).putPes(newWeight);
      }	   
      else {
	yices_retract(ctx,v[i]);
	mapConstratintsStore.erase(mapConstratintsStore.find(v[i]));	  
      }
    }
  }
  if (minor!=LONG_MAX) //if minor==LONG_MAX then there are not soft constraints in the core
    if (eqOne) atEqOne(vecProtect);
    else yices_assert(ctx,yices_mk_eq(ctx,yices_mk_sum(ctx,& vecProtect[0],vecProtect.size()),yices_mk_num(ctx,1)));
  return minor;
}
 











void WCSP::atEqOne(vector<yices_expr> & vecProtect) {
  atMostOne(vecProtect);
  atLeastOne(vecProtect);
}

void WCSP::atMostOne(vector<yices_expr> & vecProtect) {
  for (int i=0;i<vecProtect.size();i++)
    for (int j=i+1;j<vecProtect.size();j++) {
      yices_expr args[2];
      args[0]=yices_mk_not(ctx,vecProtect[i]);
      args[1]=yices_mk_not(ctx,vecProtect[j]);
      yices_assert(ctx,yices_mk_or(ctx,args,2));
    }
}

void WCSP::atLeastOne(vector<yices_expr> & vecProtect) {
  yices_assert(ctx,yices_mk_or(ctx,& vecProtect[0],vecProtect.size()));
}


///////////////////////////////////////////
//yices
///////////////////////////////////////////

lbool WCSP::yices() {
  lbool lb;
  optimtotal=0;
  if (contWeithedYices>0) {
    lb=yices_max_sat(ctx);
    if (lb==l_true) optim=yices_get_cost(yices_get_model(ctx));
  } 
  else lb=yices_check(ctx); //There are not soft constraints
  if (lb==l_true) {
    yices_model m = yices_get_model(ctx);
    yices_get_int_value(m,yices_get_var_decl_from_name(ctx,"objective"),&optimtotal);
    cout << "c ub "<< optimtotal << endl; 
    if (verbose) escriure_starts(m);
  }
  return lb;
}


void WCSP::addConstraintYices(yices_expr e,long w) {
  yices_assert_weighted(ctx,e,w);
  contWeithedYices++;
}

///////////////////////////////////////////
//linealup
///////////////////////////////////////////

lbool WCSP::linealup(bool model,int vMin,int vMax,map<string,yices_var_decl> & mapDeclVar) { 
   optim=0;
  optimtotal=vMin;
  lbool status = l_true; //yices_check(ctx);
  //if(status == l_false) return l_false;
  yices_var_decl objD=yices_get_var_decl_from_name(ctx,"objective");
  yices_expr objDecl=yices_mk_var_from_decl(ctx,objD);
  long op=vMax;
  //yices_model m = yices_get_model(ctx);
  //yices_get_int_value(m,objD,&op);
  optim=op;
  if (optim==vMin) return l_true;


  long upper_bound=op;
  long lower_bound=vMin-1;
  cout << "c lb "<< lower_bound+initialCost << endl;   
  cout << "c ub "<< upper_bound+initialCost << endl;   
  
  do {
     assertion_id idy;
    yices_assert(ctx,yices_mk_le(ctx,objDecl,yices_mk_num(ctx, upper_bound-1)));
    status = yices_check(ctx);
    if (status == l_true){
      upper_bound--;
      yices_model m = yices_get_model(ctx);
      cout << "c ub "<< upper_bound << endl;
      if (verbose) escriure_starts(m);
    } else {
      lower_bound=upper_bound-1;
      cout << "c lb "<< lower_bound << endl;
    }
  }while(status==l_true && upper_bound>vMin);
  optimtotal=upper_bound;
  status=l_true;
  return status; 
}


///////////////////////////////////////////
//linealdown
///////////////////////////////////////////

lbool WCSP::linealdown(bool model,int vMin,int vMax,map<string,yices_var_decl> & mapDeclVar) { 
  optim=0;
  optimtotal=vMin;
  lbool status = yices_check(ctx);
  if(status == l_false) return l_false;
  yices_var_decl objD=yices_get_var_decl_from_name(ctx,"objective");
  yices_expr objDecl=yices_mk_var_from_decl(ctx,objD);
  long op;
  yices_model m = yices_get_model(ctx);
  yices_get_int_value(m,objD,&op);
  optim=op;
  if (optim==vMin) return l_true;


  long upper_bound=op;
  long lower_bound=vMin-1;
  cout << "c lb "<< lower_bound+initialCost << endl;   
  cout << "c ub "<< upper_bound+initialCost << endl;   
  
  do {
    assertion_id idy=yices_assert_retractable(ctx,yices_mk_le(ctx,objDecl,yices_mk_num(ctx, lower_bound+1)));
    status = yices_check(ctx);
    if (status == l_true){
      upper_bound = lower_bound+1;
      cout << "c ub "<< upper_bound << endl;
      yices_model m = yices_get_model(ctx);
      if (verbose) escriure_starts(m);
    } else if (status == l_false){
      lower_bound++;
      yices_retract(ctx,idy);
      yices_assert(ctx,yices_mk_ge(ctx,objDecl,yices_mk_num(ctx, lower_bound+1)));
      cout << "c lb "<< lower_bound+initialCost << endl;
    }
  }while(status==l_false && lower_bound+1<upper_bound);
  optimtotal=upper_bound;
  status=l_true;
  return status; 
}


///////////////////////////////////////////
//standard
///////////////////////////////////////////

lbool WCSP::standard(bool model,int vMin,int vMax,map<string,yices_var_decl> & mapDeclVar) { 
  if (d_debug) cout<<"Iniciem Standard.."<<endl<<endl;
  optim=0;
  optimtotal=vMin;
  lbool status=l_true;
  if (rcpspmax) status=yices_check(ctx);
  if(status == l_false) return l_false;
  if (d_debug) cout<<"despres status....."<<endl<<endl;
  yices_var_decl objD=yices_get_var_decl_from_name(ctx,"objective");
  yices_expr objE = yices_mk_var_from_decl(ctx,objD);
  long op=vMax;
  if (rcpspmax) {
  if (d_debug) cout << " nooo " << endl;
    yices_model m = yices_get_model(ctx);
    yices_get_int_value(m,objD,&op);
  }
  optim=op;
  if (optim==vMin){
    y->recupera();
    return l_true;
  }
  if (d_debug) cout<<"Abans Assert.."<<op<<endl<<endl;
  yices_assert(ctx,yices_mk_le(ctx,objE,yices_mk_num(ctx,optim)));
  //yices_assert(ctx,yices_mk_le(ctx,mapDeclVar["objective"],yices_mk_num(ctx,optim)));
  if (d_debug) cout<<"Despres Assert.."<<endl<<endl;
  long upper_bound=op;
  long lower_bound=vMin-1;
  cout << "c lb "<< lower_bound+initialCost << endl;   
  cout << "c ub "<< upper_bound+initialCost << endl;   
  
  do {
    long bound;
    if ((upper_bound-lower_bound)>5) bound = (lower_bound+upper_bound)/2; else bound = upper_bound-1;
    optim=upper_bound;   
    assertion_id objId = yices_assert_retractable(ctx,yices_mk_le(ctx,objE,yices_mk_num(ctx, bound)));
    //assertion_id objId = yices_assert_retractable(ctx,yices_mk_le(ctx,mapDeclVar["objective"],yices_mk_num(ctx, bound)));
    status = yices_check(ctx);
    if(status == l_false) {
      cout << "c lb "<< bound << endl;   
      lower_bound = bound;
      yices_retract(ctx, objId);  
     }else if (status == l_true){
      y->recupera();
      upper_bound = bound;
      yices_model m = yices_get_model(ctx);
      //if (model) yices_display_model(m);
      //pausa();              
      yices_get_int_value(m,objD,&upper_bound);
      cout << "c ub "<< upper_bound<<endl;  
      if (verbose) escriure_starts(m);
    } else {cout << "error unknown" << endl;exit(1);}
  }while((lower_bound + 1)  < upper_bound );
  optimtotal=upper_bound;
  status=l_true;
  return status; 
}







///////////////////////////////////////////
//standard2
///////////////////////////////////////////

lbool WCSP::standard2(bool model,int vMin,int vMax,map<string,yices_var_decl> & mapDeclVar) { 
  optim=0;
  optimtotal=vMin;
  lbool status = yices_check(ctx);
  if(status == l_false) return l_false;
  yices_var_decl objD=yices_get_var_decl_from_name(ctx,"objective");
  yices_expr objDecl=yices_mk_var_from_decl(ctx,objD);
  long op;
  yices_model m = yices_get_model(ctx);
  yices_get_int_value(m,objD,&op);
  optim=op;
  if (optim==vMin) return l_true;
  yices_assert(ctx,yices_mk_le(ctx,objDecl,yices_mk_num(ctx,optim)));
  long upper_bound=op;
  long lower_bound=vMin-1;
  cout << "c lb "<< lower_bound+initialCost << endl;   
  cout << "c ub "<< upper_bound+initialCost << endl;   
  
 



  long powtwo=1;
  do {
    assertion_id vup_id = yices_assert_retractable(ctx,yices_mk_le(ctx,objDecl,yices_mk_num(ctx, powtwo+vMin-1)));
    status = yices_check(ctx);
    if (status==l_false) {
      yices_retract(ctx, vup_id);
      yices_assert(ctx,yices_mk_ge(ctx,objDecl,yices_mk_num(ctx, powtwo+vMin)));
      lower_bound=powtwo+vMin-1;
      cout << "c lb "<< lower_bound << endl;
      powtwo=powtwo*2;
    } else {
      upper_bound = powtwo+vMin-1;
      m = yices_get_model(ctx);
      yices_get_int_value(m,objD,&upper_bound);
      cout << "c ub "<< upper_bound << endl;
      if (verbose) escriure_starts(m);	
    }
  }while(status==l_false && (powtwo+vMin-1)<upper_bound);   

  op=upper_bound;
  
  do {
    long bound = (lower_bound+upper_bound)/2;
    optim=upper_bound;   
    assertion_id objId = yices_assert_retractable(ctx,yices_mk_le(ctx,objDecl,yices_mk_num(ctx, bound)));
    status = yices_check(ctx);
    if(status == l_false) {
      cout << "c lb "<< bound+initialCost << endl;   
      lower_bound = bound;
      yices_retract(ctx, objId);  
      yices_assert(ctx,yices_mk_ge(ctx,objDecl,yices_mk_num(ctx, bound+1)));  
    }else if (status == l_true){
      upper_bound = bound;
      yices_model m = yices_get_model(ctx);
      yices_get_int_value(m,objD,&upper_bound);
      cout << "c ub "<< upper_bound+initialCost << endl;
      if (verbose) escriure_starts(m);
    } else {cout << "error" << endl;exit(1);}
  }while((lower_bound + 1)  < upper_bound );
  optimtotal=upper_bound;
  status=l_true;
  return status; 
}


  






///////////////////////////////////////////
//common
///////////////////////////////////////////

yices_expr WCSP::reifVariable(yices_expr eb,long i) {
  //set the integer variable bi
  char aux[20];

  sprintf(aux,"bi_%i",i);
  yices_expr auxn;
  auxn=yices_mk_var_from_decl(ctx,yices_mk_var_decl(ctx,aux,yices_mk_type(ctx,"int")));
  yices_assert(ctx,yices_mk_le(ctx,auxn,yices_mk_num(ctx,1)));
  yices_assert(ctx,yices_mk_ge(ctx,auxn,yices_mk_num(ctx,0)));
  yices_expr imp1[2],imp2[2];
  imp1[0]=yices_mk_not(ctx,eb);
  imp1[1]=yices_mk_ge(ctx,auxn,yices_mk_num(ctx,1));
  yices_assert(ctx,yices_mk_or(ctx,imp1,2));
  imp2[0]=eb;
  imp2[1]=yices_mk_le(ctx,auxn,yices_mk_num(ctx,0));
  yices_assert(ctx,yices_mk_or(ctx,imp2,2));
  return auxn;
}





void WCSP::escriure_starts(yices_model m) {
  
   for (int i=0;i<=ntasques+1;i++){
        char aa[20];
        sprintf(aa,"s_%d",i);  
        long s=0;
        yices_var_decl o=yices_get_var_decl_from_name(ctx,aa);
        yices_get_int_value(m,o,&s);
        cout << "S_"<< i << "=" << s << "; ";
   }
   //yices_display_model(m);
   cout << endl;

}